"use strict";
/**
 * @license
 * Copyright 2019 Google LLC. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
Object.defineProperty(exports, "__esModule", { value: true });
/// <reference types="@webgpu/types" />
require("./flags_webgpu");
const tfjs_core_1 = require("@tensorflow/tfjs-core");
const buffer_manager_1 = require("./buffer_manager");
const argminmax_webgpu_1 = require("./kernels/argminmax_webgpu");
const binary_op = require("./kernels/binary_op_webgpu");
const binary_op_webgpu_1 = require("./kernels/binary_op_webgpu");
const concat_webgpu_1 = require("./kernels/concat_webgpu");
const conv2d_mm_webgpu_1 = require("./kernels/conv2d_mm_webgpu");
const conv2d_naive_webgpu_1 = require("./kernels/conv2d_naive_webgpu");
const depthwise_conv2d_webgpu_1 = require("./kernels/depthwise_conv2d_webgpu");
const matmul_packed_webgpu_1 = require("./kernels/matmul_packed_webgpu");
const matmul_webgpu_1 = require("./kernels/matmul_webgpu");
const maxpool_webgpu_1 = require("./kernels/maxpool_webgpu");
const pad_webgpu_1 = require("./kernels/pad_webgpu");
const resize_bilinear_webgpu_1 = require("./kernels/resize_bilinear_webgpu");
const transpose_webgpu_1 = require("./kernels/transpose_webgpu");
const unary_op = require("./kernels/unary_op_webgpu");
const unary_op_webgpu_1 = require("./kernels/unary_op_webgpu");
const webgpu_program = require("./kernels/webgpu_program");
const DEFAULT_GPUBUFFER_USAGE = GPUBufferUsage.STORAGE | GPUBufferUsage.COPY_SRC | GPUBufferUsage.COPY_DST;
class WebGPUBackend extends tfjs_core_1.KernelBackend {
    constructor(device, shaderc) {
        super();
        this.commandQueueOwnedIds = new WeakSet();
        this.tensorMap = new WeakMap();
        this.disposalQueue = [];
        this.disposed = false;
        this.uploadWaitMs = 0;
        this.downloadWaitMs = 0;
        this.binaryCache = {};
        this.device = device;
        this.queue = device.getQueue();
        this.commandQueue = [];
        this.shaderc = shaderc;
        this.compiler = new shaderc.Compiler();
        const opts = new shaderc.CompileOptions();
        opts.SetOptimizationLevel(shaderc.optimization_level.performance);
        this.compileOpts = opts;
        this.bufferManager = new buffer_manager_1.BufferManager(this.device);
    }
    floatPrecision() {
        return 32;
    }
    setDataMover(dataMover) {
        // TODO: tfjs team to implement this. Call GPUBuffer.destroy()
    }
    flushDisposalQueue() {
        this.disposalQueue.forEach(d => {
            this.releaseBuffer(d.buffer, d.byteSize, d.usage);
        });
        this.disposalQueue = [];
    }
    disposeData(dataId) {
        if (!this.tensorMap.has(dataId)) {
            throw new Error(`Tensor ${dataId} was not registered!`);
        }
        const info = this.tensorMap.get(dataId);
        if (this.commandQueueOwnedIds.has(dataId)) {
            this.disposalQueue.push(info.bufferInfo);
        }
        else {
            this.releaseBuffer(info.bufferInfo.buffer, info.bufferInfo.byteSize, info.bufferInfo.usage);
        }
        this.tensorMap.delete(dataId);
    }
    memory() {
        return {
            numBytesInGPU: this.bufferManager.numBytesUsed,
            unreliable: false
        };
    }
    getBufferManager() {
        return this.bufferManager;
    }
    acquireBuffer(byteSize, usage = DEFAULT_GPUBUFFER_USAGE) {
        return this.bufferManager.acquireBuffer(byteSize, usage);
    }
    releaseBuffer(buffer, byteSize, usage) {
        this.bufferManager.releaseBuffer(buffer, byteSize, usage);
    }
    register(dataId, shape, dtype) {
        if (!this.tensorMap.has(dataId)) {
            const byteSize = tfjs_core_1.util.sizeFromShape(shape) * tfjs_core_1.util.bytesPerElement(dtype);
            const buffer = this.acquireBuffer(byteSize);
            this.tensorMap.set(dataId, {
                values: null,
                id: -1,
                dtype,
                bufferInfo: { byteSize, usage: DEFAULT_GPUBUFFER_USAGE, buffer }
            });
        }
    }
    write(dataId, values) {
        if (!this.tensorMap.has(dataId)) {
            throw new Error(`Tensor ${dataId} was not registered!`);
        }
        const info = this.tensorMap.get(dataId);
        info.values = values;
        info.bufferInfo.buffer.setSubData(0, values);
        this.tensorMap.set(dataId, info);
    }
    submitQueue() {
        this.queue.submit(this.commandQueue.map(enc => enc.finish()));
        this.commandQueue = [];
        this.commandQueueOwnedIds = new WeakSet();
        this.flushDisposalQueue();
    }
    async getBufferData(info) {
        const staging = this.acquireBuffer(info.bufferInfo.byteSize, GPUBufferUsage.COPY_DST | GPUBufferUsage.MAP_READ);
        const encoder = this.device.createCommandEncoder({});
        encoder.copyBufferToBuffer(info.bufferInfo.buffer, 0, staging, 0, info.bufferInfo.byteSize);
        this.commandQueue.push(encoder);
        this.submitQueue();
        const mapped = await staging.mapReadAsync();
        const values = mapped.slice(0);
        staging.unmap();
        this.releaseBuffer(staging, info.bufferInfo.byteSize, GPUBufferUsage.COPY_DST | GPUBufferUsage.MAP_READ);
        return values;
    }
    convertAndCacheOnCPU(dataId, data) {
        const info = this.tensorMap.get(dataId);
        info.values = data;
        return info.values;
    }
    // TODO: Remove once this is fixed:
    // https://github.com/tensorflow/tfjs/issues/1595
    readSync(dataId) {
        const texData = this.tensorMap.get(dataId);
        const { values } = texData;
        if (values == null) {
            throw new Error('WebGPU readSync is only available for CPU-resident tensors.');
        }
        return values;
    }
    async read(dataId) {
        if (!this.tensorMap.has(dataId)) {
            throw new Error(`Tensor ${dataId} was not registered!`);
        }
        const info = this.tensorMap.get(dataId);
        const data = await this.getBufferData(info);
        const dataAsTypedArray = info.dtype === 'int32' ? new Int32Array(data) : new Float32Array(data);
        this.convertAndCacheOnCPU(dataId, dataAsTypedArray);
        return dataAsTypedArray;
    }
    async time(f) {
        const oldActiveTimers = this.activeTimers;
        const newActiveTimers = [];
        let outerMostTime = false;
        if (this.programTimersStack == null) {
            this.programTimersStack = newActiveTimers;
            outerMostTime = true;
        }
        else {
            this.activeTimers.push(newActiveTimers);
        }
        this.activeTimers = newActiveTimers;
        f();
        const flattenedActiveTimerQueries = tfjs_core_1.util.flatten(this.activeTimers.map((d) => d.query))
            .filter(d => d != null);
        const flattenedActiveTimerNames = tfjs_core_1.util.flatten(this.activeTimers.map((d) => d.name))
            .filter(d => d != null);
        this.activeTimers = oldActiveTimers;
        if (outerMostTime) {
            this.programTimersStack = null;
        }
        const kernelMs = await Promise.all(flattenedActiveTimerQueries);
        const res = {
            uploadWaitMs: this.uploadWaitMs,
            downloadWaitMs: this.downloadWaitMs,
            kernelMs: tfjs_core_1.util.sum(kernelMs),
            getExtraProfileInfo: () => kernelMs.map((d, i) => ({ name: flattenedActiveTimerNames[i], ms: d }))
                .map(d => `${d.name}: ${d.ms}`)
                .join(', '),
            wallMs: null
        };
        this.uploadWaitMs = 0;
        this.downloadWaitMs = 0;
        return res;
    }
    getAndSavePipeline(key, getBinary) {
        if (!(key in this.binaryCache)) {
            this.binaryCache[key] = getBinary();
        }
        return this.binaryCache[key];
    }
    makeOutputArray(shape, dtype) {
        return tfjs_core_1.Tensor.make(shape, {}, dtype, this);
    }
    tensorToBinding(tensor) {
        if (!tensor) {
            return null;
        }
        const tensorData = this.tensorMap.get(tensor.dataId);
        return {
            resource: {
                offset: 0,
                size: tensor.size * tfjs_core_1.util.bytesPerElement(tensor.dtype),
                buffer: tensorData.bufferInfo.buffer
            }
        };
    }
    startTimer() {
        return { startMs: tfjs_core_1.util.now(), endMs: 0 };
    }
    endTimer(query) {
        query.endMs = tfjs_core_1.util.now();
        return query;
    }
    async getQueryTime(query) {
        const timerQuery = query;
        return timerQuery.endMs - timerQuery.startMs;
    }
    compileAndRun(program, inputs, output, programUniforms) {
        if (output == null) {
            output = this.makeOutputArray(program.outputShape, inputs[0].dtype);
        }
        let dimUniforms = [];
        const bufferShapes = inputs.concat(output).map(d => d.shape);
        let currentOffset = 0;
        bufferShapes.forEach((d, i) => {
            // Uniforms.
            if (d.length === 0) {
                d = [1];
            }
            // Complete std140 layout rules are documented here:
            // tslint:disable-next-line:max-line-length
            // https://www.khronos.org/registry/OpenGL/specs/gl/glspec45.core.pdf#page=159
            let baseAlignment;
            switch (d.length) {
                case 0:
                    baseAlignment = 1;
                    break;
                case 1:
                    baseAlignment = 1;
                    break;
                case 2:
                    baseAlignment = 2;
                    break;
                case 3:
                    baseAlignment = 4;
                    break;
                case 4:
                    baseAlignment = 4;
                    break;
                default:
                    tfjs_core_1.util.assert(false, () => `Unsupported ${d.length}D shape`);
            }
            const padding = Math.ceil(currentOffset / baseAlignment) * baseAlignment -
                currentOffset;
            for (let p = 0; p < padding; ++p) {
                dimUniforms.push(0);
            }
            dimUniforms.push(...d);
            currentOffset += d.length + padding;
        });
        // TODO: handle padding of program-specific uniforms
        if (programUniforms) {
            dimUniforms = dimUniforms.concat(programUniforms);
        }
        const uniformData = new Int32Array(dimUniforms);
        const uniforms = this.makeUniforms(uniformData);
        const key = webgpu_program.makeShaderKey(program, bufferShapes.map(d => d.length));
        const inputsData = inputs.map((input, i) => ({
            // Returning dtype from tensorMap because it reflects dtype
            // of underlying buffer, rather than abstract dtype.
            dtype: this.tensorMap.get(input.dataId).dtype,
            shape: input.shape,
            name: program.variableNames[i]
        }));
        const { bindGroupLayout, pipeline } = this.getAndSavePipeline(key, () => {
            return webgpu_program.compileProgram(this.compiler, this.shaderc.shader_kind.compute, this.compileOpts, this.device, program, inputsData, output, uniforms);
        });
        const shouldTimeProgram = this.activeTimers != null;
        let query;
        if (shouldTimeProgram) {
            query = this.startTimer();
        }
        // Creating bind groups on the fly should never be a bottleneck.
        const bg = webgpu_program.makeBindGroup(this.device, bindGroupLayout, inputs.map(t => this.tensorToBinding(t)), this.tensorToBinding(output), uniforms);
        const encoder = this.device.createCommandEncoder({});
        const pass = encoder.beginComputePass();
        pass.setPipeline(pipeline);
        pass.setBindGroup(0, bg);
        pass.dispatch(program.dispatch[0], program.dispatch[1], program.dispatch[2]);
        pass.endPass();
        this.commandQueue.push(encoder);
        inputs.forEach(input => {
            this.commandQueueOwnedIds.add(input.dataId);
        });
        this.commandQueueOwnedIds.add(output.dataId);
        if (tfjs_core_1.ENV.get('WEBGPU_IMMEDIATE_EXECUTION_ENABLED')) {
            this.submitQueue();
        }
        this.releaseBuffer(uniforms.resource.buffer, uniformData.byteLength, GPUBufferUsage.COPY_DST | GPUBufferUsage.UNIFORM);
        if (shouldTimeProgram) {
            query = this.endTimer(query);
            this.activeTimers.push({ name: program.constructor.name, query: this.getQueryTime(query) });
        }
        return output;
    }
    makeUniforms(data) {
        const dimensionsBuffer = this.acquireBuffer(data.byteLength, GPUBufferUsage.COPY_DST | GPUBufferUsage.UNIFORM);
        dimensionsBuffer.setSubData(0, data);
        return {
            resource: { offset: 0, size: data.byteLength, buffer: dimensionsBuffer }
        };
    }
    pad(x, paddings, constantValue) {
        const program = new pad_webgpu_1.PadProgram(x.shape, paddings, constantValue);
        const output = this.makeOutputArray(program.outputShape, x.dtype);
        return this.compileAndRun(program, [x], output);
    }
    maxPool(x, convInfo) {
        const program = new maxpool_webgpu_1.MaxPoolProgram(convInfo);
        const output = this.makeOutputArray(program.outputShape, x.dtype);
        const dimensions = [
            convInfo.padInfo.left, convInfo.padInfo.top,
            convInfo.strideWidth, convInfo.strideHeight,
            convInfo.dilationWidth, convInfo.dilationHeight,
            convInfo.inWidth, convInfo.inHeight,
            convInfo.effectiveFilterWidth,
            convInfo.effectiveFilterHeight // Filter dims.
        ];
        return this.compileAndRun(program, [x], output, dimensions);
    }
    binaryOp(a, b, op) {
        const dtype = tfjs_core_1.backend_util.upcastType(a.dtype, b.dtype);
        const program = new binary_op_webgpu_1.BinaryOpProgram(op, a.shape, b.shape);
        const output = tfjs_core_1.Tensor.make(program.outputShape, {}, dtype);
        return this.compileAndRun(program, [a, b], output);
    }
    add(a, b) {
        return this.binaryOp(a, b, binary_op.ADD);
    }
    subtract(a, b) {
        return this.binaryOp(a, b, binary_op.SUB);
    }
    conv2d(x, filter, convInfo) {
        const output = tfjs_core_1.Tensor.make(convInfo.outShape, {}, x.dtype, this);
        let program;
        const workPerThread = tfjs_core_1.ENV.get('WEBGPU_CONV2D_WORK_PER_THREAD');
        if (workPerThread === -1) {
            // TODO(kainino0x): This may be obsolete, but is kept for reference.
            program = new conv2d_naive_webgpu_1.Conv2DNaiveProgram(convInfo);
        }
        else {
            program = new conv2d_mm_webgpu_1.Conv2DMMProgram(convInfo, workPerThread);
        }
        const pad = convInfo.padInfo.type === 'VALID' ?
            [0, 0] :
            convInfo.padInfo.type === 'SAME' ?
                [
                    -Math.floor((convInfo.filterShape[0] - 1) / 2),
                    -Math.floor((convInfo.filterShape[1] - 1) / 2)
                ] :
                [convInfo.padInfo.top, convInfo.padInfo.left];
        const dimensions = [
            convInfo.filterHeight, convInfo.filterWidth, ...pad,
            convInfo.strideHeight, convInfo.strideWidth
        ];
        return this.compileAndRun(program, [x, filter], output, dimensions);
    }
    depthwiseConv2D(x, filter, convInfo) {
        const program = new depthwise_conv2d_webgpu_1.DepthwiseConv2DProgram(convInfo);
        return this.compileAndRun(program, [x, filter]);
    }
    argMinMaxReduce(x, axis, reduceType) {
        const program = new argminmax_webgpu_1.ArgMinMaxProgram(x.shape, axis, reduceType);
        const output = this.makeOutputArray(program.outputShape, 'int32');
        return this.compileAndRun(program, [x], output, [axis]);
    }
    argMin(x, axis) {
        return this.argMinMaxReduce(x, axis, 'min');
    }
    argMax(x, axis) {
        return this.argMinMaxReduce(x, axis, 'max');
    }
    concat(tensors, axis) {
        if (tensors.length === 1) {
            return tensors[0];
        }
        // Is there a maximum number of buffers that can be uploaded to a WebGPU
        // program?
        // if (tensors.length > MAX_SSBOS_FOR_WEBGPU_PROGRAM) {
        //   const midIndex = Math.floor(tensors.length / 2);
        //   const leftSide = this.concat(tensors.slice(0, midIndex), axis);
        //   const rightSide = this.concat(tensors.slice(midIndex), axis);
        //   return this.concat([leftSide, rightSide], axis);
        // }
        const outShape = tfjs_core_1.backend_util.computeOutShape(tensors.map(t => t.shape), axis);
        const tensors2D = tensors.map(t => t.reshape([
            tfjs_core_1.util.sizeFromShape(t.shape.slice(0, axis)),
            tfjs_core_1.util.sizeFromShape(t.shape.slice(axis))
        ]));
        const program = new concat_webgpu_1.ConcatProgram(tensors2D.map(t => t.shape));
        const res = this.compileAndRun(program, tensors2D);
        return res.reshape(outShape);
    }
    multiply(a, b) {
        return this.binaryOp(a, b, binary_op.MUL);
    }
    floorDiv(a, b) {
        return this.binaryOp(a, b, binary_op.INT_DIV);
    }
    sigmoid(x) {
        const program = new unary_op_webgpu_1.UnaryOpProgram(x.shape, unary_op.SIGMOID);
        return this.compileAndRun(program, [x]);
    }
    relu(x) {
        const program = new unary_op_webgpu_1.UnaryOpProgram(x.shape, unary_op.RELU);
        return this.compileAndRun(program, [x]);
    }
    resizeBilinear(x, newHeight, newWidth, alignCorners) {
        const program = new resize_bilinear_webgpu_1.ResizeBilinearProgram(x.shape, newHeight, newWidth, alignCorners);
        const output = this.makeOutputArray(program.outputShape, x.dtype);
        return this.compileAndRun(program, [x], output);
    }
    reshape(x, shape) {
        return tfjs_core_1.Tensor.make(shape, { dataId: x.dataId }, x.dtype);
    }
    cast(x, dtype) {
        return tfjs_core_1.backend_util.castTensor(x, dtype, this);
    }
    transpose(x, perm) {
        const program = new transpose_webgpu_1.TransposeProgram(x.shape, perm);
        return this.compileAndRun(program, [x]);
    }
    batchMatMul(a, b, transposeA, transposeB) {
        // TODO: Support transposed inputs.
        // const outerShapeA = transposeA ? a.shape[2] : a.shape[1];
        // const outerShapeB = transposeB ? b.shape[1] : b.shape[2];
        const outerShapeA = a.shape[1];
        const outerShapeB = b.shape[2];
        const [batch, ,] = a.shape;
        const output = tfjs_core_1.Tensor.make([batch, outerShapeA, outerShapeB], {}, a.dtype, this);
        let program;
        // TODO: We should eventually use the blocked version, but keeping around
        // the old version while we try to understand conditions under which blocked
        // is faster.
        if (tfjs_core_1.ENV.get('WEBGPU_MATMUL_WORK_PER_THREAD') === 0) {
            program = new matmul_webgpu_1.MatMulProgram(output.shape);
        }
        else {
            program = new matmul_packed_webgpu_1.MatMulPackedProgram(output.shape, tfjs_core_1.ENV.get('WEBGPU_MATMUL_WORK_PER_THREAD'));
        }
        return this.compileAndRun(program, [a, b], output);
    }
    fromPixels(pixels, numChannels) {
        if (pixels == null) {
            throw new Error('pixels passed to tf.browser.fromPixels() can not be null');
        }
        const outShape = [pixels.height, pixels.width, numChannels];
        let imageData = pixels.data;
        if (tfjs_core_1.ENV.getBool('IS_BROWSER')) {
            if (!(pixels instanceof HTMLVideoElement) &&
                !(pixels instanceof HTMLImageElement) &&
                !(pixels instanceof HTMLCanvasElement) &&
                !(pixels instanceof ImageData) &&
                !(pixels.data instanceof Uint8Array)) {
                throw new Error('pixels passed to tf.browser.fromPixels() must be either an ' +
                    `HTMLVideoElement, HTMLImageElement, HTMLCanvasElement, ImageData` +
                    ` or {data: Uint32Array, width: number, height: number}, ` +
                    `but was ${pixels.constructor.name}`);
            }
            if (pixels instanceof HTMLVideoElement ||
                pixels instanceof HTMLImageElement ||
                pixels instanceof HTMLCanvasElement) {
                if (this.fromPixels2DContext == null) {
                    this.fromPixels2DContext =
                        document.createElement('canvas').getContext('2d');
                    this.fromPixels2DContext.canvas.width = pixels.width;
                    this.fromPixels2DContext.canvas.height = pixels.height;
                }
                this.fromPixels2DContext.drawImage(pixels, 0, 0, pixels.width, pixels.height);
                pixels = this.fromPixels2DContext.canvas;
            }
            // TODO: Remove this once we figure out how to upload textures directly to
            // WebGPU.
            const imageDataLivesOnGPU = pixels instanceof HTMLVideoElement ||
                pixels instanceof HTMLImageElement ||
                pixels instanceof HTMLCanvasElement;
            if (imageDataLivesOnGPU) {
                imageData = this.fromPixels2DContext
                    .getImageData(0, 0, pixels.width, pixels.height)
                    .data;
            }
        }
        // TODO: Encoding should happen on GPU once we no longer have to download
        // image data to the CPU.
        let pixelArray = imageData;
        if (numChannels != null && numChannels !== 4) {
            pixelArray = new Uint8Array(pixels.width * pixels.height * numChannels);
            for (let i = 0; i < imageData.length; i++) {
                if (i % 4 < numChannels) {
                    const pixelIndex = Math.floor(i / 4);
                    pixelArray[pixelIndex * numChannels + i % 4] = imageData[i];
                }
            }
        }
        const output = this.makeOutputArray(outShape, 'int32');
        this.write(output.dataId, Int32Array.from(pixelArray));
        return output;
    }
    dispose() {
        if (this.disposed) {
            return;
        }
        this.bufferManager.dispose();
        this.disposed = true;
    }
}
exports.WebGPUBackend = WebGPUBackend;
//# sourceMappingURL=backend_webgpu.js.map