"use strict";
/**
 * @license
 * Copyright 2019 Google LLC. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
Object.defineProperty(exports, "__esModule", { value: true });
const tfjs_core_1 = require("@tensorflow/tfjs-core");
const webgpu_util_1 = require("../webgpu_util");
const matmul_packed_webgpu_1 = require("./matmul_packed_webgpu");
const matmul_webgpu_1 = require("./matmul_webgpu");
class Conv2DMMProgram {
    constructor(convInfo, workPerThread) {
        this.variableNames = ['x', 'W'];
        this.uniforms = 'ivec2 filterDims, pad, stride;';
        this.workGroupSize = [
            16, 16,
            1
        ];
        this.outputShape = convInfo.outShape;
        tfjs_core_1.util.assert(convInfo.dataFormat === 'channelsLast', () => 'TODO: NCHW is unimplemented');
        tfjs_core_1.util.assert(convInfo.dilationHeight === 1 && convInfo.dilationWidth === 1, () => 'TODO: Dilation is unimplemented');
        let elementsPerThread;
        let matMulSource;
        if (workPerThread === 0) {
            elementsPerThread = [1, 1, 1];
            matMulSource = matmul_webgpu_1.makeMatMulSource();
        }
        else {
            elementsPerThread = [workPerThread, workPerThread, 1];
            matMulSource = matmul_packed_webgpu_1.makeMatMulPackedSource(workPerThread);
        }
        this.dispatchLayout = { x: [1], y: [2], z: [0] };
        const matMulOutShape = [
            convInfo.outShape[0], convInfo.outShape[1] * convInfo.outShape[2],
            convInfo.outShape[3]
        ];
        this.dispatch = webgpu_util_1.computeDispatch(this.dispatchLayout, matMulOutShape, this.workGroupSize, elementsPerThread);
        this.userCode = `
        ${matMulSource}

        int batch;

        float mm_readA(uint row, uint col) {
          int r = int(row), c = int(col);
          ivec4 coord = ivec4(
              (c / filterDims[1]) % filterDims[0],
              c % filterDims[1],
              c / (filterDims[0] * filterDims[1]),
              r);

          ivec4 shape = ivec4(filterDims, xShape[3], outShape[3]);
          return coordIsValid(coord, shape) ? W[getFlatIndex(coord, shape)] : 0;
        }

        float mm_readB(uint row, uint col) {
          int r = int(row), c = int(col);
          int outRow = c / outShape[2];
          int outCol = c % outShape[2];

          int WRow = (r / filterDims[1]) % filterDims[0];
          int WCol = r % filterDims[1];

          ivec4 coord = ivec4(
              batch,
              pad[0] + outRow * stride[0] + WRow,
              pad[1] + outCol * stride[1] + WCol,
              r / (filterDims[0] * filterDims[1]));
          return coordIsValid(coord, xShape) ?
              x[getFlatIndex(coord, xShape)] : 0;
        }

        void mm_write(uint row, uint col, float value) {
          ivec4 outCoord = ivec4(
              batch,
              col / outShape[2],
              col % outShape[2],
              row);
          if (coordIsValid(outCoord, outShape)) {
            result[getFlatIndex(outCoord, outShape)] = value;
          }
        }

        void main() {
          batch = int(gl_GlobalInvocationID.z);

          int dimAOuter = outShape[3];
          int dimBOuter = outShape[1] * outShape[2];
          int dimInner = filterDims[0] * filterDims[1] * xShape[3];
          mm_matMul(dimAOuter, dimInner, dimBOuter);
        }
      `;
    }
}
exports.Conv2DMMProgram = Conv2DMMProgram;
//# sourceMappingURL=conv2d_mm_webgpu.js.map